import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import tdsbrondata
from pyspark.sql import functions as F
from pyspark.sql.types import *
from delta.tables import DeltaTable
from functools import reduce

itemsRequiredColumns = {
    "DienstId": IntegerType(),
    "Source": StringType(),
    "SurrogateKey": LongType(),
    "FacturatieMaand": StringType(),
    "VerkooprelatieId": StringType(),
    "ItemCode": StringType(),
    "Aantal": FloatType()
}

itemsOptionalColumns = {
    "AfwijkendePrijs": FloatType(),
    "DatumVanOrigineel": StringType(),
    "DatumTotOrigineel": StringType(),
    "AantalOrigineel": FloatType(),
    "DurationOrigineel": FloatType(),
    "EntityType": StringType(),
    "EntityId": StringType(),
    "EntityLabel": StringType()
}

def validateItemsSchema(items):
    schemaDict = {f.name: f.dataType for f in items.schema.fields}

    for columnName, expectedType in itemsRequiredColumns.items():
        if columnName not in schemaDict:
            raise ValueError(f"Missing required column '{columnName}'")
        if type(schemaDict[columnName]) != type(expectedType):
            raise TypeError(f"Column '{columnName}' has type {schemaDict[columnName]}, expected {expectedType}")

    for columnName, expectedType in itemsOptionalColumns.items():
        if columnName in schemaDict and type(schemaDict[columnName]) != type(expectedType):
            raise TypeError(f"Column '{columnName}' has type {schemaDict[columnName]}, expected {expectedType}")

def writeItems(items):

    # TEMP: Later voor waarschuwen
    items = items.filter(F.col("VerkooprelatieId").isNotNull())
    items = items.filter(F.col("ItemCode").isNotNull())

    items = items.withColumn("ItemId", F.expr("uuid()"))
    items = writeLines(items)

    table = "items"
    tablePath = f"{tdsbrondata.tablesRootPath}/{table}"

    if DeltaTable.isDeltaTable(tdsbrondata._spark, tablePath):
        deltaTable = DeltaTable.forPath(tdsbrondata._spark, tablePath)
        facturatieMaand = items.select("FacturatieMaand").head()["FacturatieMaand"]
        deltaTable.delete(f"FacturatieMaand = '{facturatieMaand}'")
        items.write.format("delta").mode("append").save(tablePath)
    else:
        items.write.format("delta").mode("overwrite").option("overwriteSchema", "true").save(tablePath)

def writeLines(items):

    table = "lines"
    tablePath = f"{tdsbrondata.tablesRootPath}/{table}"
    facturatieMaand = items.select("FacturatieMaand").head()["FacturatieMaand"]

    # Split into items to preserve and to items aggregate

    existingItemsOptionalColumns = [col for col in itemsOptionalColumns.keys() if col in items.columns]

    if existingItemsOptionalColumns:
        filterExpression = " AND ".join([f"{col} IS NULL" for col in existingItemsOptionalColumns])
        itemsToPreserve = items.filter(f"NOT ({filterExpression})")
        itemsToAggregate = items.filter(filterExpression)
    else:
        itemsToPreserve = items.limit(0)
        itemsToAggregate = items

    for col, colType in itemsOptionalColumns.items():
        if col not in items.columns:
            itemsToAggregate = itemsToAggregate.withColumn(col, F.lit(None).cast(colType))
            itemsToPreserve = itemsToPreserve.withColumn(col, F.lit(None).cast(colType))

    # Make lines for items to preserve

    itemsToPreserve = itemsToPreserve.withColumn("LineId", F.expr("uuid()")).cache()
    linesFromItemsToPreserve = itemsToPreserve.drop("ItemId")
    linesFromItemsToPreserve = linesFromItemsToPreserve.withColumn(
        "HasItems",
        F.when(F.col("SurrogateKey").isNotNull(), True).otherwise(False)
    ).drop("Source", "SurrogateKey", "EntityType", "EntityId", "EntityLabel")

    # Make lines for items to aggregate

    linesFromItemsToAggregate = (
        itemsToAggregate
        .groupBy("DienstId", "FacturatieMaand", "VerkooprelatieId", "ItemCode")
        .agg(
            F.sum("Aantal").alias("Aantal"),
            F.collect_list("SurrogateKey").alias("SurrogateKeys")
        )
        .withColumn("LineId", F.expr("uuid()"))
    )

    for col, colType in itemsOptionalColumns.items():
        if col not in linesFromItemsToAggregate.columns:
            linesFromItemsToAggregate = linesFromItemsToAggregate.withColumn(col, F.lit(None).cast(colType))

    linesFromItemsToAggregate = linesFromItemsToAggregate.withColumn(
        "HasItems",
        F.expr("size(SurrogateKeys) > 0")
    ).drop("Source", "SurrogateKey", "EntityType", "EntityId", "EntityLabel")

    # Separate lines with and without items

    linesWithItems = linesFromItemsToAggregate.filter(F.col("HasItems") == True)
    linesWithoutItems = linesFromItemsToAggregate.filter(F.col("HasItems") == False)

    # Attach items to lines that have items

    if linesWithItems.count() > 0:
        itemsAttachedToLines = (
            linesWithItems
            .select(
                "LineId",
                F.explode("SurrogateKeys").alias("SurrogateKey"),
                "DienstId",
                "FacturatieMaand",
                "VerkooprelatieId",
                "ItemCode"
            )
            .join(
                itemsToAggregate,
                on=[
                    "SurrogateKey",
                    "DienstId",
                    "FacturatieMaand",
                    "VerkooprelatieId",
                    "ItemCode"
                ],
                how="left"
            )
        )
    else:
        itemsAttachedToLines = tdsbrondata._spark.createDataFrame([], schema=itemsToAggregate.schema)
        itemsAttachedToLines = itemsAttachedToLines.withColumn("LineId", F.lit(None).cast(StringType()))

    # Identify items not attached to any line

    attachedItemIds = itemsAttachedToLines.select("ItemId").distinct()
    itemsNotAttachedToLines = itemsToAggregate.join(attachedItemIds, on="ItemId", how="left_anti")
    itemsNotAttachedToLines = itemsNotAttachedToLines.withColumn("LineId", F.lit(None).cast(StringType()))

    # Combine

    itemsToAggregate = itemsAttachedToLines.unionByName(itemsNotAttachedToLines)
    linesFromItemsToAggregate = linesFromItemsToAggregate.drop("SurrogateKeys")

    # Final

    items = itemsToPreserve.unionByName(itemsToAggregate)
    items = items.withColumnRenamed("ItemId", "Id")

    lines = linesFromItemsToPreserve.unionByName(linesFromItemsToAggregate)
    lines = lines.withColumnRenamed("LineId", "Id")

    if DeltaTable.isDeltaTable(tdsbrondata._spark, tablePath):
        deltaTable = DeltaTable.forPath(tdsbrondata._spark, tablePath)
        deltaTable.delete(f"FacturatieMaand = '{facturatieMaand}'")
        lines.write.format("delta").mode("append").save(tablePath)
    else:
        lines.write.format("delta").mode("overwrite").option("overwriteSchema", "true").save(tablePath)

    return items

def mergeItems(lists, facMonth):

    items = []

    schemas = [row.NotebookName for row in lists.select("NotebookName").distinct().collect()]
    for schema in schemas:
        
        df = tdsbrondata.utils.getCurrentData(
            workspaceName='Tosch-Facturatie',
            lakehouseName='B_Datawarehouse',
            schemaName=schema,
            tableName='items',
            usesScd=False
        )

        df = df.filter(F.col("FacturatieMaand") == facMonth)
        
        items.append(df)

    dfItems = reduce(lambda df1, df2: df1.unionByName(df2), items)

    return dfItems

def mergeLines(lists, facMonth):

    lines = []

    schemas = [row.NotebookName for row in lists.select("NotebookName").distinct().collect()]
    for schema in schemas:
        df = tdsbrondata.utils.getCurrentData(
            workspaceName='Tosch-Facturatie',
            lakehouseName='B_Datawarehouse',
            schemaName=schema,
            tableName='lines',
            usesScd=False
        )

        df = df.filter(F.col("FacturatieMaand") == facMonth)
        
        lines.append(df)

    dfLines = reduce(lambda df1, df2: df1.unionByName(df2), lines)

    return dfLines




import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import tdsbrondata
from pyspark.sql import functions as F
from pyspark.sql.column import Column
from pyspark.sql.types import *

def getSecret(secretName):
    return tdsbrondata._notebookutils.credentials.getSecret(tdsbrondata.keyvaultUrl, secretName)

def getCurrentData(workspaceName, lakehouseName, schemaName, tableName, usesScd=False, scdMoment=None):
    
    abfss = f"abfss://{workspaceName}@onelake.dfs.fabric.microsoft.com/{lakehouseName}.Lakehouse/Tables/{schemaName}/{tableName}"

    data = tdsbrondata._spark.read.format("delta").load(abfss)
    
    if usesScd:
        if scdMoment is not None:
            data = data.filter(
                (F.col("ScdStartDate") <= scdMoment) &
                (
                    F.col("ScdEndDate").isNull() |
                    (F.col("ScdEndDate") > scdMoment)
                )
            )
        else:
            data = data.filter(
                (F.col("CurrentFlag") == 1) &
                (F.col("ScdEndDate").isNull())
            )

    return data

def getCurrentDataSQL(schemaName, tableName, usesScd=False, scdMoment=None):

    table = f"{schemaName}.{tableName}"
    whereClause = ""

    if isinstance(scdMoment, Column):
        ts_value = (
            tdsbrondata._spark
                .range(1)
                .select(scdMoment.alias("ts"))
                .collect()[0]["ts"]
        )

        scdMomentLiteral = f"'{ts_value}'"
    else:
        scdMomentLiteral = None

    if usesScd:
        if scdMomentLiteral is not None:
            whereClause = f"""
            WHERE ScdStartDate <= {scdMomentLiteral}
              AND (ScdEndDate IS NULL OR ScdEndDate > {scdMomentLiteral})
            """
        else:
            whereClause = """
            WHERE CurrentFlag = 1
              AND ScdEndDate IS NULL
            """

    query = f"(SELECT * FROM {table} {whereClause}) AS t"

    return (
        tdsbrondata._spark.read
            .format("jdbc")
            .option("url", tdsbrondata.jdbcUrl)
            .option("dbtable", query)
            .option("accessToken", tdsbrondata.token)
            .load()
    )

def getAvailableSourceFiles(schemaName, tableName, sourceDataMode, start, directory="", lakehouseName="B_Diensten"):

    availableSourceFiles = []

    if sourceDataMode == 'automatic' and directory == "":
        sourceDataPathRoot = f"abfss://Tosch-Data@onelake.dfs.fabric.microsoft.com/{lakehouseName}.Lakehouse/Files/AutomaticData/{schemaName}/"
        for item in tdsbrondata._notebookutils.fs.ls(sourceDataPathRoot):
            if item.isDir:
                folder = item.name.rstrip("/")
                if folder.isdigit() and int(folder) >= int(start):
                    folderPath = f"{sourceDataPathRoot}{folder}/"
                    try:
                        files = tdsbrondata._notebookutils.fs.ls(folderPath)
                        for f in files:
                            fileName = f.name.rstrip("/")
                            if fileName == f"{tableName}_full.parquet":
                                filePath = f"{folderPath}{fileName}"
                                ScdMoment = folder[:4] + "-" + folder[4:6] + "-" + folder[6:8] + " 00:00:00"
                                availableSourceFiles.append({"filePath": filePath, "ScdMoment": ScdMoment})
                    except Exception as e:
                        print(f"Cannot list files in folder: {folderPath}, {e}")
    elif sourceDataMode == 'automatic':
        sourceDataPathRoot = f"abfss://Tosch-Data@onelake.dfs.fabric.microsoft.com/{lakehouseName}.Lakehouse/Files/AutomaticData/{schemaName}/{directory}/"
        files = tdsbrondata._notebookutils.fs.ls(sourceDataPathRoot)
        for f in files:
            fileName = f.name.rstrip("/")
            prefix = fileName.split("_")[0]
            if prefix.isdigit() and int(prefix) >= int(start) and fileName.endswith(".csv"):
                filePath = f"{sourceDataPathRoot}{fileName}"
                ScdMoment = prefix[:4] + "-" + prefix[4:6] + "-01 00:00:00"
                availableSourceFiles.append({"filePath": filePath, "ScdMoment": ScdMoment})
    else:
        sourceDataPathRoot = f"abfss://Tosch-Data@onelake.dfs.fabric.microsoft.com/{lakehouseName}.Lakehouse/Files/ManualData/{schemaName}/"
        files = tdsbrondata._notebookutils.fs.ls(sourceDataPathRoot)
        for f in files:
            fileName = f.name.rstrip("/")
            prefix = fileName.split("_")[0]
            if prefix.isdigit() and int(prefix) >= int(start) and tableName in fileName and fileName.endswith(".csv"):
                filePath = f"{sourceDataPathRoot}{fileName}"
                ScdMoment = prefix[:4] + "-" + prefix[4:6] + "-01 00:00:00"
                availableSourceFiles.append({"filePath": filePath, "ScdMoment": ScdMoment})

    return availableSourceFiles

def addEventLogEntry(lakehouseName, schemaName, tableName, eventType, eventResult):

    abfss = 'abfss://Tosch-Data@onelake.dfs.fabric.microsoft.com/M_Metadata.Lakehouse/Tables/dbo'
    
    schema = StructType([
        StructField("LakehouseName", StringType(), True),
        StructField("SchemaName", StringType(), True),
        StructField("TableName", StringType(), True),
        StructField("EventType", StringType(), True),
        StructField("EventResult", StringType(), True)
    ])

    row = {"LakehouseName": lakehouseName, "SchemaName": schemaName, "TableName": tableName, "EventType": eventType, "EventResult": eventResult}
    df = tdsbrondata._spark.createDataFrame([row], schema=schema)

    try:
        dfExisting = tdsbrondata._spark.read.format("delta").load(f"{abfss}/eventlog")
    except Exception as e:
        dfExisting = tdsbrondata._spark.createDataFrame([], schema=schema)

    isDuplicate = (
        dfExisting.filter(
            (F.col("LakehouseName") == row["LakehouseName"]) &
            (F.col("SchemaName") == row["SchemaName"]) &
            (F.col("TableName") == row["TableName"]) &
            (F.col("EventType") == row["EventType"]) &
            (F.col("EventResult") == row["EventResult"])
        )
        .limit(1)
        .count() > 0
    )

    if not isDuplicate:
        df.write.mode("append").format("delta").save(f"{abfss}/eventlog")

def fileToBeProcessed(lakehouseName, schemaName, tableName, filePattern):

    shouldBeProcessed = True

    eventLog = tdsbrondata._spark.read.format("delta").load("abfss://Tosch-Data@onelake.dfs.fabric.microsoft.com/M_Metadata.Lakehouse/Tables/dbo/eventlog")

    fileAdded = eventLog.filter(
        (F.col("LakehouseName") == lakehouseName) & 
        (F.col("SchemaName") == schemaName) & 
        (F.col("TableName") == tableName) & 
        (F.col("EventType") == "fileAdded") & 
        (F.col("EventResult").startswith(filePattern))
    ).limit(1)

    fileAdded = fileAdded.collect()
    if fileAdded:
        fileName = fileAdded[0]["EventResult"]
        fileProcessed = eventLog.filter(
            (F.col("LakehouseName") == lakehouseName) & 
            (F.col("SchemaName") == schemaName) & 
            (F.col("TableName") == tableName) & 
            (F.col("EventType") == "fileProcessed") & 
            (F.col("EventResult").startswith(filePattern))
        ).limit(1)

        fileProcessed = fileProcessed.collect()
        if fileProcessed:
            shouldBeProcessed = False
    else:
        shouldBeProcessed = False

    if shouldBeProcessed:
        return fileName
    else:
        return None